## this version of the script includes threading for data collection and spawns a pool of processes for report generation
## this script is not published in a functional form; it is a legacy script which has been replaced by OSRSBuildDatabase.py (https://poignanttech.com/projects/)

import sqlite3
import csv
import glob
from glob import glob; from os.path import expanduser
import pandas as pd
from functools import reduce
import requests
import json
import time
import os
import sys
import logging
from logging.handlers import RotatingFileHandler
from io import StringIO
from queue import Queue
import threading
from threading import Thread
import multiprocessing as mp
import numpy as np

##change working directory to script location
pathtorunfile = os.path.dirname(__file__)
os.chdir(pathtorunfile)

##configure logging
#logging.basicConfig(level=logging.DEBUG, filename="logfile.txt", filemode="a+", format="%(asctime)-15s %(levelname)-8s %(message)s")
log_formatter = logging.Formatter('%(asctime)s %(levelname)s %(funcName)s(%(lineno)d) %(message)s')
logfile = 'logfile.txt'
my_handler = RotatingFileHandler(logfile, mode='a', maxBytes=1024*1024, backupCount=2, encoding=None, delay=0)
my_handler.setFormatter(log_formatter)
my_handler.setLevel(logging.INFO)
app_log = logging.getLogger('root')
app_log.setLevel(logging.INFO)
app_log.addHandler(my_handler)

##define global variables
#output directory for all reports
csvoutputdir = 'csvoutput/'
#define urlprefix and headers, both used for API calls
urlprefix = 'https://prices.runescape.wiki/api/v1/osrs/'
#headers = {
#'User-Agent': 'enter contact information here',
}
#define the endpoints that we want check and update reports for
EndpointList = ['24h', '1h', '5m']
#define blank list to be used to contain endpoints which need reports updated
EndpointReportGenList = []
#define the number of processes we want to spawn when processing data for each endpoint
cpucount = 1

### define report parameters
#GenerateReport set to 'Yes' will force the creation of new reports regardless if new data has been gathered
#values in ReportIntervals dictate how many endpoint iterations starting from the present will be used to generate a report, named with each corresponding ReportName. The quantity of ReportIntervals must match the quantity of ReportNames
def reportparameters(CurrentEndpoint):
	if CurrentEndpoint == '24h':
		CurrentReportDict = {
		"csvdir": "dailymarketdatacsvs/",
		"timeinterval": 86400,
		"intervalname": "Daily",
		"GenerateReport": "",
		"ReportIntervals": [7, 30, 360],
		"ReportNames": ["WeeklyReport", "MonthlyReport", "YearlyReport",]
		}
	if CurrentEndpoint =='1h':
		CurrentReportDict = {
		"csvdir": "hourlymarketdatacsvs/",
		"timeinterval": 3600,
		"intervalname": "Hourly",
		"GenerateReport": "",
		"ReportIntervals": [24, 360, 720],
		"ReportNames": ["GranularDailyReport", "GranularBiweeklyReport", "GranularMonthlyReport"]
		}
	if CurrentEndpoint =='5m':
		CurrentReportDict = {
		"csvdir": "fiveminutemarketdatacsvs/",
		"timeinterval": 300,
		"intervalname": "FiveMinute",
		"GenerateReport": "",
		"ReportIntervals": [1, 12, 288],
		"ReportNames": ["VeryGranularFiveMinuteReport", "VeryGranularHourlyReport", "VeryGranularDailyReport"]
		}
	return CurrentReportDict

### define JSON to CSV conversion commands
#the following conversions will be used against every JSON file that we download
def jsoncsvcode(url, csvdir):
	jsonresponse = requests.get(url, headers=headers)
	jsondata = jsonresponse.json()
	currenttimestamp = jsondata["timestamp"]
	responsedata = jsondata["data"]
	keylist = list(responsedata.keys())
	valuelist = list(responsedata.values())
	totalvalues = []
	totalkeysandvalues = []
	for value in valuelist:
		subvalues = list(value.values())
		totalvalues.append(subvalues)
	for key1, value1 in zip(keylist, totalvalues):
		totalkeysandvalues.append([key1] + value1)
	with open(f'{csvdir}{currenttimestamp}.csv', 'w', newline='') as f:
		writer = csv.writer(f, quoting=csv.QUOTE_MINIMAL, quotechar="'")
		writer.writerow(['id','avgHighPrice','highPriceVolume','avgLowPrice','lowPriceVolume'])
		writer.writerows(totalkeysandvalues)
	#the API is rate limited and will return null values if we query more than once per second, so we add a 4-second pause here
	time.sleep(4)
	return currenttimestamp

### define data collection routine
#this routine checks the age of existing data for the endpoints that we wish to generate reports for and downloads/converts any missing data into csv.
def DataCollect(CurrentEndpoint):
	app_log.info(f'checking existing {CurrentEndpoint} data')
	#set our report parameters
	CurrentReportDict = reportparameters(CurrentEndpoint)
	#pull values and lists from our report parameters
	csvdir = CurrentReportDict["csvdir"]
	timeinterval = CurrentReportDict["timeinterval"]
	intervalname = CurrentReportDict["intervalname"]
	GenerateReport = CurrentReportDict["GenerateReport"]
	ReportIntervals = CurrentReportDict["ReportIntervals"]
	ReportNames = CurrentReportDict["ReportNames"]
	EndpointReportGenList = []
	#check if output directory exists
	if os.path.exists(csvoutputdir) is False:
		app_log.info('CSV Report Directory does not exist: Creating now')
		os.makedirs(csvoutputdir)
	#check if market data location exists
	if os.path.exists(csvdir) is False:
		app_log.info('CSV Data Directory does not exist: Creating now')
		os.makedirs(csvdir)
	#get the current time and subtract the timestamp from the timestamp of our most recent .csv file to determine if we need to pull new data
	CurrentTime = time.time()
	if len(os.listdir(csvdir)) != 0:
		newest_csv_file = sorted(glob(os.path.join(csvdir, "*.csv")), key=os.path.basename, reverse = True)[0]
		newest_csv_file_trunc = int(newest_csv_file.replace('.csv', '').replace(f'{csvdir}', ''))
		app_log.info(f'{CurrentEndpoint} data is {((CurrentTime - newest_csv_file_trunc) - timeinterval) / 3600} hours old')
	#if there is no csv file, set the recording var to 0 which will trigger pulling new data
	else:
		newest_csv_file_trunc = 0
	#obtain new data if most recent data is old. NOTE: We multiply the timeinterval by 2 because when these endpoints are updated, their new timestamps do not reflect the update time, but rather the update time - timeinterval
	if (CurrentTime - newest_csv_file_trunc) > (timeinterval * 2):
		app_log.info(f'Current {CurrentEndpoint} data out-of-date: Downloading latest file')
		#since there is new data, we will want to generate fresh reports
		GenerateReport = 'Yes'
		#api call, put json into string and execute json conversion to csv subroutine
		url = urlprefix + CurrentEndpoint
		#jsoncsvcode performs json query, converts and saves json data to csv, and returns the current timestamp
		currenttimestamp = jsoncsvcode(url, csvdir)
		#create/update symbolic link (used so that it is easy for client reporting software to access recent data)
		#if os.path.islink(f'{CurrentEndpoint}.csv') is True:
		#	os.unlink(f'{CurrentEndpoint}.csv')
		#os.symlink(f'{csvdir}{currenttimestamp}.csv', f'{CurrentEndpoint}.csv')
		app_log.info(f'Download of latest {CurrentEndpoint} data completed; now checking for missing endpoint data required for our report intervals')
	else:
		app_log.info(f'Current {CurrentEndpoint} data is up-to-date; now checking for missing endpoint data required for our report intervals')
		currenttimestamp = newest_csv_file_trunc
	#check against the csvdir and fetch any missing historical data, starting from the oldest data defined as the product of our report's time interval and the quantity used in our largest report, iterating toward and stopping at the present timestamp
	iterativetimestamp = currenttimestamp - (timeinterval * ReportIntervals[-1:][0])
	while iterativetimestamp != currenttimestamp:
		if os.path.isfile(f'{csvdir}{iterativetimestamp}.csv') is False:
			#since there was missing data, we will want to generate fresh reports
			GenerateReport = 'Yes'
			url = f'{urlprefix}{CurrentEndpoint}?timestamp={iterativetimestamp}'
			#call the jsoncsvcode subroutine again to grab the data and save as csv; we don't collect the timestamp this time since we already have it
			jsoncsvcode(url, csvdir)
			app_log.info(f'Download and CSV conversion completed for historical {intervalname} timestamp: {iterativetimestamp}')		
		iterativetimestamp = iterativetimestamp + timeinterval
	app_log.info(f'Historical {CurrentEndpoint} timestamp data acquisition completed')
	##Generate a new report if new data has been captured or if reports have been manually set to run, otherwise skip
	if GenerateReport == 'Yes':
		ReportIntervalMax = max(ReportIntervals)
		nameinterval = 0
		for ReportInterval in ReportIntervals:
			EndpointReportGenList = EndpointReportGenList + [[CurrentEndpoint, ReportInterval, csvdir, ReportNames[nameinterval], ReportIntervalMax, timeinterval, currenttimestamp,]]
			nameinterval = nameinterval + 1
		return EndpointReportGenList
	else:
		return 'null'

### define data importing logic
# this routine grabs data that we need to use for all reports and concatenates the appropriate data for each report into pandas dataframes.
# each resulting database is concated with report generation information and passed into the report generation queue
def DataImport(EndpointReportGenList):
	#start timer
	dataimportstart = time.time()
	#pull our variables out of the list of reports we need to generate
	EndpointIntervals = ([(item[0], item[2], item[4]) for item in EndpointReportGenList])
	#this defines the scope of data we need to collect (endpoint name and interval) and where we can access the data
	UniqueEndpointIntervals = (set(EndpointIntervals))
	#list used to contain dataframes extracted from each csv file for a given endpoint
	current_dataframe_list = []
	for UniqueEndpointInterval in UniqueEndpointIntervals:
		CurrentEndpoint = UniqueEndpointInterval[0]
		csvdir = UniqueEndpointInterval[1]
		ReportInterval = UniqueEndpointInterval[2]
		#create empty dataframe
		all_endpoint_dataframes = pd.DataFrame(columns=['id', 'avgHighPrice', 'highPriceVolume', 'avgLowPrice', 'lowPriceVolume'])
		all_endpoint_dataframes.set_index(['id'])
		#obtain list of csv files from specified directory
		csv_files = sorted(glob(os.path.join(csvdir, "*.csv")), key=os.path.basename, reverse = True)
		app_log.info(f'Iterating over {ReportInterval} report intervals')
		#iterate over each csv file for the current endpoint, qty limited by the report interval for our largest report
		for f in csv_files[:ReportInterval]:
			#import csv file as pandas dataframe
			current_dataframe = pd.read_csv(f, sep=",")
			#add dataframe to list of dataframes
			current_dataframe_list = current_dataframe_list + [current_dataframe]
	#... concate our list of dataframes and delete the list
	EndpointReportGenListWithData = []
	for EndpointReportGen in EndpointReportGenList:
		#since current_dataframe_list is sorted chronologically, we can select the slice of data we need for the target report
		dataframe_concat = pd.concat(current_dataframe_list[0:(EndpointReportGen[1])])
		EndpointReportGen = EndpointReportGen + [dataframe_concat]
		EndpointReportGenListWithData = EndpointReportGenListWithData + [EndpointReportGen]
	return(EndpointReportGenListWithData)
	
### define report generation logic
# this function pulls data from the ReportGenQueue, generating and exporting a csv file for any single report we wish to generate, extracting the necessary data from the pandas dataframe generated from the DataImport function.
# this routine spawns child processes of the ReportGenWorker function. The quantity of child processes depends on the cpucount global variable.
# only one instance of this function may run at a time, as this is where our worker pools are created and destroyed
def ReportGen(Report):
	## Extract items needed to generate a report
	ReportInterval = Report[1]
	ReportName = Report[3]
	PandasDB = Report[7]
	## Pre-processing
	#sort pandas dataframe by id. This ensures that datapoints pertaining to each unique ID do not get separated from each other if we are using multiprocessing
	PandasDB.sort_values(by='id', axis=0, inplace=True)
	#copy id which we will use for determining "DatapointQuantity" and "DatapointMax" values. If we don't do this before setting an index, the groupby command will throw a fit because there will be an "id" index and an "id" column.
	PandasDB['idCopy'] = PandasDB.loc[:, 'id']
	#index pandas dataframe. We are creating an index here because in my testing this yielded a performance improvement.
	PandasDB.set_index(keys=['id'], drop=False,inplace=True)
	#count the qty of datapoints available for each unique id, "DatapointQuantity" and drop idCopy...
	PandasDB['DatapointQuantity'] = PandasDB.groupby('idCopy')['idCopy'].transform('count')
	PandasDB = PandasDB.drop(['idCopy'], axis=1)	
	#... Define "DatapointMax" column based off of our current ReportInterval...
	PandasDB['DatapointMax'] = ReportInterval
	#... and finally, pipe a list of unique IDs to a python variable "currentids"
	currentids=PandasDB['id'].unique().tolist()
	## Split workload, for multiprocessing
	#split list of ids into as many arrays as we have CPUs defined and add them to a list "currentidsarrays"
	currentidsarrays = np.array_split(currentids, cpucount)
	#likewise, split our pandas dataframe into as many dataframes as we have arrays of ids and add them to a list "split_dataframes"
	split_dataframes = []
	for currentidsarray in currentidsarrays:
		startingid = int(currentidsarray[0])
		endingid = int(currentidsarray[-1])
		split_dataframe = PandasDB.loc[(PandasDB.id >= startingid) & (PandasDB.id <= endingid)]
		split_dataframes = split_dataframes + [[currentidsarray] + [split_dataframe]]
	## Processing	
	#spawn a worker pool, providing each worker with a corresponding split dataframe, with results being placed in a list "dataframe_list"
	with mp.Pool(cpucount) as p:
		dataframe_list = p.map(ReportGenWorker, split_dataframes)
	## Post-Processing
	#concatenate dataframe_list so that we can combine the work performed by the worker pool
	dataframe_concat = pd.concat(dataframe_list)
	#fix the index in-case things came back in the wrong order
	dataframe_concat.sort_values('id')
	#export to CSV
	dataframe_concat.to_csv(f'{csvoutputdir}{ReportName}.csv', sep=',', index=False, quoting=csv.QUOTE_NONNUMERIC, quotechar="'")
	app_log.info(f'{ReportName} Completed')

### define report calculation subprocess function
#this function performs the most computationally-intensive work of this script. It has been designed to run in-parallel with itself, with each instance performing computations of a subset of a given report's data. This data is returned to the parent ReportGen function and recombined before a report is exported
#this function can effectively scale to thousands of CPU cores, but more realistically it scales to about ~8 with the default reporting settings before we see more substantial diminishing returns.
def ReportGenWorker(values):
	#define variables
	currentidsarray = values[0]
	split_dataframe = values[1]
	##connect to "TempDB" temporary SQLite database and import the "split_dataframe" this process has been assigned to chew on
	TempDB = sqlite3.connect(":memory:")
	cur = TempDB.cursor()
	cur.executescript('''CREATE TABLE MasterTable('id' INT, 'avgHighPrice' INT,'highPriceVolume' INT,'avgLowPrice' INT,'lowPriceVolume' INT, 'DatapointQuantity' INT, 'DatapointMax' INT);''')
	cur.executescript('''CREATE TEMP TABLE FinalOutput('id' INT, 'MedianLow' INT,'MedianHigh' INT,'MedianVolumeLow' INT,'MedianVolumeHigh' INT,'MeanLow' INT,'MeanHigh' INT,'MeanVolumeLow' INT,'MeanVolumeHigh' INT,'MinLow' INT,'MinHigh' INT,'MaxLow' INT,'MaxHigh' INT);''')
	split_dataframe.to_sql('MasterTable', TempDB, if_exists="append", index=False)
	#create an index for our temp SQLite database before proceeding, which will significantly improve execution time.
	cur.execute('''CREATE INDEX indices ON MasterTable(id);''')
	for currentid in currentidsarray:
		## Create temporary table for the id currently specified within the python for-loop
		cur.execute(f'''CREATE TEMP TABLE currentid AS SELECT * FROM MasterTable WHERE id = {currentid}''')	
		## Data normalization
		# Our table will contain some null values that we need to correct. Since volume moves as a function of time, any null volume entries need to be set to 0. Since price is a function of volume, any null price values need to be set to empty strings. Performing this step is crucial as it ensures that our subsequent calculations are accurate. To provide an example, if no products are sold within a one-hour period of time, volume sold within that hour is 0. Price on the other hand cannot be said to be any value, so this datapoint should be empty/ignored.
		cur.executescript('''CREATE TEMP TABLE AvgLowPriceFixNull AS SELECT avgLowPrice, lowPriceVolume FROM currentid WHERE avgLowPrice != ''; CREATE TEMP TABLE AvgHighPriceFixNull AS SELECT avgHighPrice, highPriceVolume FROM currentid WHERE avgHighPrice != ''; CREATE TEMP TABLE LowPriceVolumeFixNull AS SELECT lowPriceVolume FROM currentid; UPDATE LowPriceVolumeFixNull SET lowPriceVolume = 0 WHERE lowPriceVolume = ''; CREATE TEMP TABLE HighPriceVolumeFixNull AS SELECT highPriceVolume FROM currentid; UPDATE HighPriceVolumeFixNull SET highPriceVolume = 0 WHERE highPriceVolume = '';''')
		# Our table may also be missing complete rows of datapoints for some IDs within our reporting window. If an item experiences no Buy or Sell trade volume within a set span of time, the OSRS Wiki API will simply not return any values at all, resulting in an absense of row(s) for that ID. While this is fine in the context of price, which is not a direct function of time, this is extremely problematic in the context of volume, because any calculations involving ids with missing datapoints will be skewed higher than they should be, since we will be missing ALL rows where High and Low volume are both equal to "0". This is a significant factor for "Granular" (1h) and "VeryGranular" (5m) reports where there can often be zero volume bought/sold within these shorter spans of time for any given id.
		# While "DatapointMax" reflects the maximum amount of datapoints a single id can have, "DatapointQuantity" reflects the quantity of datapoints actually provided by the API. The difference between these reflect how many "0" entries we need to stuff into our High and Low volume tables before we can proceed with calculations.
		cur.execute('''CREATE TEMP TABLE Datapoints AS SELECT DISTINCT (DatapointMax - DatapointQuantity) AS MissingDatapointQuantity FROM currentid;''')
		# Pipe "MissingDatapointQuantity" into a Python integer so we can execute an if loop. Ideally I would do this entirely within SQLite... maybe I'll revisit this at some point but for now it isn't a priority because it "just werks" doesn't seem to add significant overhead... It just looks a little ugly and out of place in a block that is otherwise mostly SQLite...
		MissingDatapoints = cur.execute("SELECT MissingDatapointQuantity FROM Datapoints;").fetchall()
		for MissingDatapoint in MissingDatapoints:
			MissingDatapoint = int(MissingDatapoint[0])
			# If an item has missing datapoints, we will use overloaded multiplication in-order to specify and pipe the desired quantity of "0" value rows into a single SQLite command
			if MissingDatapoint != 0:
				MissingRows = ('(0),' * MissingDatapoint)
				#remove trailing comma from our string
				MissingRows = MissingRows[:-1]
				cur.executescript(f'''INSERT INTO LowPriceVolumeFixNull VALUES {MissingRows}; INSERT INTO HighPriceVolumeFixNull VALUES {MissingRows};''')
		## Data Calculations
		# Mean calculations for price. While we could use the native "avg" function here, we can obtain a more accurate mean by weighing each price against its associated volume for each datapoint before calculating a mean value. This will prevent our mean from disproportionately skewing high or low in circumstances where small quantities of volume have traded at prices much higher or lower than what is typically encountered.
		cur.executescript('''CREATE TEMP TABLE TotalLowVol AS SELECT sum(lowPriceVolume) AS TotalLowVol FROM AvgLowPriceFixNull; CREATE TEMP TABLE TotalHighVol AS SELECT sum(highPriceVolume) AS TotalHighVol FROM AvgHighPriceFixNull;''')
		cur.executescript('''CREATE TEMP TABLE LowPriceVolumeProduct AS SELECT (avgLowPrice * lowPriceVolume) AS LowPriceVolumeProduct FROM AvgLowPriceFixNull; CREATE TEMP TABLE HighPriceVolumeProduct AS SELECT (avgHighPrice * highPriceVolume) AS HighPriceVolumeProduct FROM AvgHighPriceFixNull;''')
		cur.executescript('''CREATE TEMP TABLE MeanLowTable AS SELECT (sum(LowPriceVolumeProduct.LowPriceVolumeProduct)) / TotalLowVol.TotalLowVol AS MeanLow FROM TotalLowVol, LowPriceVolumeProduct; CREATE TEMP TABLE MeanHighTable AS SELECT (sum(HighPriceVolumeProduct.HighPriceVolumeProduct)) / TotalHighVol.TotalHighVol AS MeanHigh FROM TotalHighVol, HighPriceVolumeProduct;''')
		# Mean calculations for volume. We can just use the avg function here because volume moves as a function of time.
		cur.executescript('''CREATE TEMP TABLE MeanLowVolumeTable AS SELECT AVG(lowPriceVolume) AS MeanVolumeLow FROM LowPriceVolumeFixNull; CREATE TEMP TABLE MeanHighVolumeTable AS SELECT AVG(highPriceVolume) AS MeanVolumeHigh FROM HighPriceVolumeFixNull;''')
		# Median calculations (this is very ugly because sqlite does not have a native median function)
		# Unlike our mean calculations for price, I am not weighing median prices against their respective volumes. Nonwithstanding how complicated these median commands already are, momentary spikes/dips in price typically do not skew median price as significantly as they do mean price when unweighted against relative volume. I also don't use median price all that much in my reporting, and if I did I could just increase granularity to 1h/5m to improve accuracy. 
		# That said, I will probably come back and improve this the next time I'm looking for a "fun" math/statistics exercise.
		cur.executescript('''CREATE TEMP TABLE MedianLowTable AS SELECT AVG(avgLowPrice) AS MedianLow FROM (SELECT avgLowPrice FROM AvgLowPriceFixNull ORDER BY avgLowPrice LIMIT 2 - (SELECT COUNT(*) FROM AvgLowPriceFixNull) % 2 OFFSET (SELECT (COUNT(*) - 1) / 2 FROM AvgLowPriceFixNull)); CREATE TEMP TABLE MedianHighTable AS SELECT AVG(avgHighPrice) AS MedianHigh FROM (SELECT avgHighPrice FROM AvgHighPriceFixNull ORDER BY avgHighPrice LIMIT 2 - (SELECT COUNT(*) FROM AvgHighPriceFixNull) % 2 OFFSET (SELECT (COUNT(*) - 1) / 2 FROM AvgHighPriceFixNull)); CREATE TEMP TABLE MedianLowVolumeTable AS SELECT AVG(lowPriceVolume) AS MedianVolumeLow FROM (SELECT lowPriceVolume FROM LowPriceVolumeFixNull ORDER BY lowPriceVolume LIMIT 2 - (SELECT COUNT(*) FROM LowPriceVolumeFixNull) % 2 OFFSET (SELECT (COUNT(*) - 1) / 2 FROM LowPriceVolumeFixNull)); CREATE TEMP TABLE MedianHighVolumeTable AS SELECT AVG(highPriceVolume) AS MedianVolumeHigh FROM (SELECT highPriceVolume FROM HighPriceVolumeFixNull ORDER BY highPriceVolume LIMIT 2 - (SELECT COUNT(*) FROM HighPriceVolumeFixNull) % 2 OFFSET (SELECT (COUNT(*) - 1) / 2 FROM HighPriceVolumeFixNull));''')
		# Min/Max calculations
		cur.executescript('''CREATE TEMP TABLE MinLowPriceTable AS SELECT min(avgLowPrice) AS MinLow FROM AvgLowPriceFixNull; CREATE TEMP TABLE MinHighPriceTable AS SELECT min(avgHighPrice) AS MinHigh FROM AvgHighPriceFixNull; CREATE TEMP TABLE MaxLowPriceTable AS SELECT max(avgLowPrice) AS MaxLow FROM AvgLowPriceFixNull; CREATE TEMP TABLE MaxHighPriceTable AS SELECT max(avgHighPrice) AS MaxHigh FROM AvgHighPriceFixNull;''')
		# Insert all of the values for the current id into the FinalOutput table
		cur.execute('''INSERT INTO FinalOutput SELECT DISTINCT currentid.id AS id, MedianLow, MedianHigh, MedianVolumeLow, MedianVolumeHigh, MeanLow, MeanHigh, MeanVolumeLow, MeanVolumeHigh, MinLow, MinHigh, MaxLow, MaxHigh FROM currentid CROSS JOIN MedianLowTable CROSS JOIN MedianHighTable CROSS JOIN MedianLowVolumeTable CROSS JOIN MedianHighVolumeTable CROSS JOIN MeanLowTable CROSS JOIN MeanHighTable CROSS JOIN MeanLowVolumeTable CROSS JOIN MeanHighVolumeTable CROSS JOIN MinLowPriceTable CROSS JOIN MinHighPriceTable CROSS JOIN MaxLowPriceTable CROSS JOIN MaxHighPriceTable;''')
		# Cleanup - purge temp tables before proceeding to next id
		cur.executescript('''DROP TABLE IF EXISTS currentid; DROP TABLE IF EXISTS Datapoints; DROP TABLE IF EXISTS AvgLowPriceFixNull; DROP TABLE IF EXISTS AvgHighPriceFixNull; DROP TABLE IF EXISTS LowPriceVolumeFixNull; DROP TABLE IF EXISTS HighPriceVolumeFixNull; DROP TABLE IF EXISTS MedianLowTable; DROP TABLE IF EXISTS MedianHighTable; DROP TABLE IF EXISTS MedianLowVolumeTable; DROP TABLE IF EXISTS MedianHighVolumeTable; DROP TABLE IF EXISTS TotalLowVol; DROP TABLE IF EXISTS TotalHighVol; DROP TABLE IF EXISTS LowPriceVolumeProduct; DROP TABLE IF EXISTS HighPriceVolumeProduct; DROP TABLE IF EXISTS MeanLowTable; DROP TABLE IF EXISTS MeanHighTable; DROP TABLE IF EXISTS MeanLowVolumeTable; DROP TABLE IF EXISTS MeanHighVolumeTable; DROP TABLE IF EXISTS MinLowPriceTable; DROP TABLE IF EXISTS MinHighPriceTable; DROP TABLE IF EXISTS MaxLowPriceTable; DROP TABLE IF EXISTS MaxHighPriceTable;''')
	# Insert all of the values for the current id into the FinalOutput table
	split_result = pd.read_sql_query("SELECT * FROM FinalOutput", TempDB)
	# Cleanup
	cur.executescript('''DROP TABLE IF EXISTS currentid; DROP TABLE IF EXISTS Datapoints; DROP TABLE IF EXISTS AvgLowPriceFixNull; DROP TABLE IF EXISTS AvgHighPriceFixNull; DROP TABLE IF EXISTS LowPriceVolumeFixNull; DROP TABLE IF EXISTS HighPriceVolumeFixNull; DROP TABLE IF EXISTS MedianLowTable; DROP TABLE IF EXISTS MedianHighTable; DROP TABLE IF EXISTS MedianLowVolumeTable; DROP TABLE IF EXISTS MedianHighVolumeTable; DROP TABLE IF EXISTS TotalLowVol; DROP TABLE IF EXISTS TotalHighVol; DROP TABLE IF EXISTS LowPriceVolumeProduct; DROP TABLE IF EXISTS HighPriceVolumeProduct; DROP TABLE IF EXISTS MeanLowTable; DROP TABLE IF EXISTS MeanHighTable; DROP TABLE IF EXISTS MeanLowVolumeTable; DROP TABLE IF EXISTS MeanHighVolumeTable; DROP TABLE IF EXISTS MinLowPriceTable; DROP TABLE IF EXISTS MinHighPriceTable; DROP TABLE IF EXISTS MaxLowPriceTable; DROP TABLE IF EXISTS MaxHighPriceTable;''')
	return(split_result)

### define itemlist collection and conversion logic
#the itemlist has a unique format and requires its own set of fetch/conversion operations
#since this is a fairly static endpoint, fetching this is skipped if the current version is under a day old
def ItemlistCollection(csvoutputdir):
	#check if output directory exists
	if os.path.exists(csvoutputdir) is False:
		app_log.info('CSV Report Directory does not exist: Creating now')
		os.makedirs(csvoutputdir)
	#check age of current itemlist
	CurrentTime = time.time()
	if os.path.exists(f'{csvoutputdir}itemlist.csv') == True:
		itemlistmtime = os.path.getmtime(f'{csvoutputdir}itemlist.csv')
		app_log.info(f'itemlist is {(CurrentTime - itemlistmtime) / 3600} hours old')
	else:
		app_log.info('itemlist does not exist; downloading now')
		itemlistmtime = 0
	#obtain new itemlist if more than a day old (ie. 86400 seconds)
	if CurrentTime - itemlistmtime > 86400:
		#change target URL. we will use the same header used earlier.
		url = 'https://oldschool.runescape.wiki/?title=Module:GEIDs/data.json&action=raw&ctype=application%2Fjson'

		#define our request and query the API
		r = requests.get(url, headers=headers)
		data = json.dumps(r.json())

		#modify the resulting string, convert JSON to csv
		data2 = data.replace(',', '\n')
		separator = '(UTC)'
		data3 = data2.split(separator, 1)[1]
		data4 = data3.replace('"\n', 'name,id\n')
		data5 = data4.replace(': ', ',')
		data6 = data5.replace(' "', '')
		data7 = data6.replace('"', '')
		data8 = data7.replace('}', '')

		#export to itemlist.csv
		with open(f'{csvoutputdir}itemlist.csv', 'w', ) as f:
			f.write(data8)

		app_log.info('item list update completed')
		#api is rate-limited; if we pull data we should sleep for at least 1 second
		time.sleep(1)
	else:
		app_log.info('item list already recent; skipping acquisition')

### define mapping collection and conversion logic
#the itemlist also has a unique format and requires its own set of fetch/conversion operations
#since this is a fairly static endpoint, fetching this is skipped if the current version is under a day old
def MappingCollection(csvoutputdir):
	#check if output directory exists
	if os.path.exists(csvoutputdir) is False:
		app_log.info('CSV Report Directory does not exist: Creating now')
		os.makedirs(csvoutputdir)
	#check age of current itemlist
	CurrentTime = time.time()
	if os.path.exists(f'{csvoutputdir}mapping.csv') == True:
		itemlistmtime = os.path.getmtime(f'{csvoutputdir}mapping.csv')
		app_log.info(f'mapping is {(CurrentTime - itemlistmtime) / 3600} hours old')
	else:
		app_log.info('mapping does not exist; downloading now')
		itemlistmtime = 0

	#obtain new itemlist if more than a day old (ie. 86400 seconds)
	if CurrentTime - itemlistmtime > 86400:
		#change target URL. we will use the same header used earlier.
		url = 'https://prices.runescape.wiki/api/v1/osrs/mapping'
		#define our request and query the API
		jsonresponse = requests.get(url, headers=headers)
		jsondata = jsonresponse.json()
		df = pd.DataFrame(jsondata)
		#this pop command drops the "examine" column which is unnecessary and creates issues due to arbitrary usage of quotation, asterisk, commas semicolons, etc...
		df.pop(df.columns[0])
		df.to_csv(f'{csvoutputdir}mapping.csv', index=False)
		app_log.info('mapping update completed')
		#api is rate-limited; if we pull data we should sleep for at least 1 second
		time.sleep(1)
	else:
		app_log.info('mapping already recent; skipping acquisition')

### define web scraping routine for black market GP exchange rate
#for our reporting, we need the black market GP exchange rate, which is not available from the OSRS Wiki API and must be scraped from gold-trading websites
#if any data is missing or malformed, procedures will be passed in-order to allow mean calculation and/or the rest of the script to proceed
#this routine is NOT included in the published blog version of the script; this is only for production.
#the output is saved to the script root rather than csvdir, since the end result will not contain IDs and will need to be concatenated with the mastertable separately
def BlackMarketScrape():
	#check age of cached black market exchange rate
	CurrentTime = time.time()
	if os.path.exists('blackmarket.csv') == True:
		blackmarketmtime = os.path.getmtime('blackmarket.csv')
		app_log.info(f'black market data is {(CurrentTime - blackmarketmtime) / 3600} hours old')
	else:
		app_log.info('black  market data does not exist; scraping now')
		blackmarketmtime = 0
	#scrape data from black market websites if existing data is more than a day old (ie. 86400 seconds)
	if CurrentTime - blackmarketmtime > 86400:
		# add some additional modules
		#import requests
		from bs4 import BeautifulSoup
		import re
		from statistics import mean
		## define fake headers
		#fakeheaders = {'User-Agent': 'enter fake headers here'}
		## define empty value list
		goldpricevalues = []
		## scrape from gold selling site #1
		#URL = "enter a gold selling site here"
		try:
			page = requests.get(URL, headers=fakeheaders)
			soup = BeautifulSoup(page.content, "html.parser")
			currentprice = # this is where you hack and slash the html to get what you're looking for
			goldpricevalues.append(currentprice)
		except:
			pass
		## export to csv if a mean can be calculated
		try:			
			with open('blackmarket.csv', 'w', newline='') as f:
				writer = csv.writer(f, quoting=csv.QUOTE_MINIMAL, quotechar="'")
				writer.writerow([mean(goldpricevalues)])
		except:
			pass
	else:
		app_log.info('black market data already recent; skipping acquisition')
		
### define routine for joining reports and creating master picklefile
#this routine combines all of the .csv files contained in the csvoutput directory, using the id column as a key
#data is exported as a .pkl pandas dataframe file which can be used for client report generation
def MasterTableCreate(csvoutputdir):
	
	#create empty dataframes list
	dataframes = []
	#iterate over all report files contained in csv
	for report in os.listdir(f'{csvoutputdir}'):
		#truncate report names so that they can be spliced onto their respective columns
		filename_trunc = report.replace('.csv', '').replace('Report', '')
		with open(f'{csvoutputdir}{report}','r') as fin:
			#grab headers from csv file, perform string replace and 
			header_line = fin.readline()
			header_line = header_line.replace("'", "").replace('\n', '')
			header_line_list = header_line.split(",")
			prepended_header_line_list = []
			for header in header_line_list:
				if header != 'id':
					header = filename_trunc + header
				prepended_header_line_list.append(header)
		currentpandasdata = pd.read_csv(f'{csvoutputdir}{report}')
		currentpandasdata.columns = prepended_header_line_list
		dataframes.append(currentpandasdata)
	#merge dataframes
	df_merged = reduce(lambda  left,right: pd.merge(left,right,on=['id'],how='outer'), dataframes)
	#merge black market exchange rate ***NOT INCLUDED IN BLOG/PUBLISHED VERSION OF THIS SCRIPT***
	with open('blackmarket.csv', 'r') as fin:
		blackmarketdata = fin.readline()
		blackmarketdata.replace('\n', '')
	df_merged['BlackMarketRate']=f'{float(blackmarketdata)}'
	#save merged dataframe to picklefile
	df_merged.to_pickle('MasterTable.pkl', compression=None, protocol=5)
	app_log.info(f'pickle file "MasterTable.pkl" creation completed')







############ Define worker functions for parallelization ###############

#this function, combined with the queue and worker daemon defined in our runtime, allows us to run the DataImport function in a separate thread
def DataImportInvoke():
	while True:
		EndpointIntervalList = DataImportQueue.get()
		EndpointIntervalListWithData = DataImport(EndpointIntervalList)
		for EndpointIntervalReport in EndpointIntervalListWithData:
			ReportGenQueue.put(EndpointIntervalReport)
		DataImportQueue.task_done()

#this function, combined with the queue and worker daemon defined in our runtime, allows us to run the ReportGen function in a separate thread
def ReportGenInvoke():
	while True:
		EndpointIntervalReport = ReportGenQueue.get()
		ReportGen(EndpointIntervalReport)
		ReportGenQueue.task_done()


########################## RUNTIME #############################

if __name__ == '__main__':
	#start our queues and worker threads/pools
	startruntime = time.time()
	DataImportQueue = Queue(maxsize = 0)
	threading.Thread(target=DataImportInvoke, daemon=True).start()
	ReportGenQueue = Queue(maxsize = 0)
	threading.Thread(target=ReportGenInvoke, daemon=True).start()

	#collect latest data for all defined endpoints
	for Endpoint in EndpointList:
		EndpointIntervalList = DataCollect(Endpoint)
		if EndpointIntervalList != 'null':
			DataImportQueue.put(EndpointIntervalList)
		
	###Collect and export itemlist
	ItemlistCollection(csvoutputdir)

	###Collect and export mapping
	MappingCollection(csvoutputdir)
	
	###Collect and export black market exchange rate
	BlackMarketScrape()
	
	#proceed once DataImportQueue is empty
	DataImportQueue.join()

	#proceed once ReportGenQueue is empty
	ReportGenQueue.join()
	
	#generate MasterTable Picklefile once all data has been collected and reports processed
	MasterTableCreate(csvoutputdir)

	app_log.info(f'total runtime: {time.time() - startruntime} seconds')
